from contextlib import ExitStack
from functools import partial
from itertools import repeat
from os import listdir, mkdir, remove
from os.path import basename, dirname, exists, join, splitext
from warnings import warn

import numpy as np
from tqdm.contrib.concurrent import process_map

from videochef.io import VideoReader, VideoWriter
from videochef.util import count_frames, make_batch_sequence, unwrap_dictionary


def parallel_proc_frame(
    vid_path,
    video_reader_kwargs,
    writer_fullfiles,
    frame_by_frame_kwarg_dicts,
    analysis_func=None,
    overwrite_tmp=True,
):
    """Helper function to pipe + process the video frames from one vid to another.

        NB: all args must be positional except analysis func for process_map to work correctly.

    Arguments:
        vid_path {str} -- path to the input video
        video_reader_kwargs {dict} -- kwargs to pass to VideoReader
        writer_fullfiles {iterable of str} -- where to write the processed data
        frame_by_frame_kwarg_dicts -- list of dictionaries of frame-by-frame kwargs (eg [{'key1': val1, 'key2': val1, ...}, {'key1': val2, 'key2': val2, ...}, ...])

    Keyword Arguments:
        analysis_func {function} -- a single processing function.
            Must accept one video frame as a sole positional arg.
            Must return a tuple of outputs to be written to writer_fullfiles.
                For videos: return a frame (if output_type is 'video')
                For arrays: return a dictionary of scalars (if output_type is 'arrays')

    Raises:
        ValueError: if no analysis function is provided.
    """

    # Check inputs
    if analysis_func is None:
        raise ValueError("Please provide an analysis function!")

    if frame_by_frame_kwarg_dicts is None:
        frame_by_frame_kwarg_dicts = repeat({})

    # Check if overwriting
    if not overwrite_tmp and any([exists(wf) for wf in writer_fullfiles]):
        print("Not overwriting")
        return

    # Parse inputs
    output_exts = [splitext(wf)[1] for wf in writer_fullfiles]
    batch_len = len(video_reader_kwargs["frame_ixs"])

    # Prep context managed objs for videos
    movie_types = [".avi", ".mp4"]
    vc_reader = VideoReader(vid_path, **video_reader_kwargs)
    vid_writer_idx = [i for i, ext in enumerate(output_exts) if ext in movie_types]
    if len(vid_writer_idx) > 0:
        vid_writer_idx = vid_writer_idx[
            0
        ]  # TODO: allow multiple output vids from a single video?
        vc_writer = VideoWriter(writer_fullfiles[vid_writer_idx])
    else:
        vc_writer = None

    # Prep npz
    npz_idx = [i for i, ext in enumerate(output_exts) if ext == ".npz"]
    if len(npz_idx) > 0:
        npz_idx = npz_idx[0]
        output = {}
    else:
        npz_idx = None

    with ExitStack() as stack:

        # Enter vid contexts
        vid = stack.enter_context(vc_reader)
        if vc_writer is not None:
            writer = stack.enter_context(vc_writer)

        # Loop through frames
        for iFrame, (frame, frame_kwarg_dict) in enumerate(
            zip(vid, frame_by_frame_kwarg_dicts)
        ):

            # Do the analysis
            out = analysis_func(frame, **frame_kwarg_dict)
            if type(out) is not tuple:
                raise ValueError(
                    f"Analysis function must return a tuple of outputs but returned type {type(out)} instead!"
                )

            # Deal with movie data
            if vc_writer is not None:
                frame_out = out[vid_writer_idx]
                writer.append(frame_out)

            # Deal with npz data
            if npz_idx is not None:
                npz_data = out[npz_idx]
                if type(npz_data) is not dict:
                    raise ValueError(
                        f"npz portion of analysis function's returned tuple must contain a dictionary but contained {type(npz_data)} instead!"
                    )
                if iFrame == 0:
                    for key in npz_data.keys():
                        try:
                            output[key] = np.zeros((batch_len, *npz_data[key].shape))
                        except MemoryError:
                            warn(
                                f"Unable to allocate memory for key {key}, skipping stitching it..."
                            )
                for key in output.keys():
                    output[key][iFrame, ...] = npz_data[key]

        # After full vid, save npz data
        if npz_idx is not None:
            np.savez(writer_fullfiles[npz_idx], **output)

    return


def video_chef(
    func,
    path_to_vid,
    func_global_kwargs=None,
    func_frame_kwargs=None,
    output_types=["video"],
    max_workers=3,
    frame_batch_size=500,
    every_nth_frame=1,
    truncate_to_n_batches=None,
    vid_read_reporter=False,
    video_reader_kwargs=None,
    tmp_dir=None,
    overwrite_tmp=True,
    remove_chunks=True,
    proc_suffix="_PROC",
):
    """Process a video in embarassingly parallel batches, writing out a processed video (an avi) or arrays of scalars (an npz).
    Thin wrapper around tqdm.contrib.concurrent.process_map.

    Arguments:
        func {python function} -- the processing function.
            Must accept one video frame as the sole positional arg.
            Must return a tuple. Tuple can contain either:
                -- a frame (if output_type is 'video')
                -- a dictionary of scalars (if output_type is 'arrays')
                -- both
        path_to_vid {str} -- path to the video to be processed.

    Keyword Arguments:
        func_global_kwargs {dict} -- kwargs to pass into the analysis function once, at initialization
        func_frame_kwargs {dict} -- kwargs to pass for each frame (eg {'key1': vals, 'key2': vals}, where vals is an array of same length as the video)
        output_types {iterable of str} -- 'video' or 'arrays' (default: {['video']}).
        max_workers {int} -- max_workers for process_map (default: {3})
        frame_batch_size {int} -- n frames processed per worker-batch (default: {500})
        every_nth_frame {int} -- process every nth frame (default: {1})
        truncate_to_n_batches {int} -- if not None, only process this many batches (for debugging) {default: None}
        vid_read_reporter {bool} -- if True, workers will report which video frames they're reading (mostly for debugging) (default: {False})
        tmp_dir {[type]} -- where to store the temporary (unstitched) processed videos (default: {path_to_vid/tmp})
        overwrite_tmp {bool} -- if True, remove anything in the tmp dir before starting; if False, try to use anything pre-existing (default: {True})
        remove_chunks {bool} -- if True, remove the temporary (unstitched) processed videos after stitching (default: {True})
        proc_suffix {str} -- suffix for the outputs (default: {'_PROC'})

    Returns:
        [str] -- path to the stitched and processed video / npz
    """

    nframes = count_frames(path_to_vid)
    batch_seq = make_batch_sequence(
        nframes, frame_batch_size, 0, offset=0, step=every_nth_frame
    )
    if truncate_to_n_batches is not None:
        batch_seq = batch_seq[:truncate_to_n_batches]
    n_expected_frames = sum([len(list(r)) for r in batch_seq])
    vid_name, vid_ext = splitext(basename(path_to_vid))
    vid_dir = dirname(path_to_vid)

    if tmp_dir is None:
        tmp_dir = join(vid_dir, "tmp")

    if not exists(tmp_dir):
        mkdir(tmp_dir)
    elif overwrite_tmp:
        for file in listdir(tmp_dir):
            remove(join(tmp_dir, file))

    if video_reader_kwargs is None:
        video_reader_kwargs = {}

    print("Processing in parallel...")

    ### Prep iters for process_map ###
    # NB, all are length of batch_seq

    # Get video writers to use  #TODO: get extension from passed in vid, so can use mp4, eg
    parallel_writer_fullfiles = []
    for iOutput, output_type in enumerate(output_types):
        if output_type == "video":
            writers_vid_idx = iOutput
            parallel_writer_fullfiles.append(
                [join(tmp_dir, f"proc_{i}{vid_ext}") for i in range(len(batch_seq))]
            )
        elif output_type == "arrays":
            writers_npz_idx = iOutput
            parallel_writer_fullfiles.append(
                [join(tmp_dir, f"proc_{i}.npz") for i in range(len(batch_seq))]
            )
        else:
            raise ValueError(f"Unrecognized output_type {output_type}")
    # Zip the writers into pairs
    parallel_writer_fullfiles = list(zip(*parallel_writer_fullfiles))

    # Set up kwargs for video readers
    if vid_read_reporter:
        reporter_vals = (i for i in range(len(batch_seq)))
    else:
        reporter_vals = (None for i in range(len(batch_seq)))
    partial_reader_kwarg_lists = [
        {
            "frame_ixs": np.array(batch),
            "reporter_val": reporter_val,
            **video_reader_kwargs,
        }
        for batch, reporter_val in zip(batch_seq, reporter_vals)
    ]

    # Get chunks of frame-by-frame kwargs if needed
    if func_frame_kwargs is None:
        func_frame_kwargs = {}
    kwarg_dict_list = np.array(unwrap_dictionary(func_frame_kwargs), dtype="object")
    if len(kwarg_dict_list) == 0:
        partial_fr_by_fr_kwarg_list = [[{} for _ in batch] for batch in batch_seq]
    elif len(kwarg_dict_list) == nframes:
        partial_fr_by_fr_kwarg_list = [kwarg_dict_list[batch] for batch in batch_seq]
    else:
        raise ValueError(
            f"Expected empty func_frame_kwargs or lists of length n_expected_frames ({n_expected_frames}), but got lists of length {len(kwarg_dict_list)}. Did you pass every_nth_frame correctly?"
        )

    ### Do the parallel processing ###

    # Apply global kwargs
    if func_global_kwargs is None:
        func_global_kwargs = {}
    analysis_func_partial = partial(func, **func_global_kwargs)

    # Get parallel proc function
    proc_func = partial(
        parallel_proc_frame,
        analysis_func=analysis_func_partial,
        overwrite_tmp=overwrite_tmp,
    )

    # Call to process_map is (func, iter1, iter2, ..., iterN, **kwargs) where iter1 - iterN are iterables that are passed as positional args to func.
    # Importantly, chunksize=1 means workers get only one item from each iterable at once; nb that each single "chunk" will be frame_batch_size frames!
    process_map(
        proc_func,
        repeat(path_to_vid),
        partial_reader_kwarg_lists,
        parallel_writer_fullfiles,
        partial_fr_by_fr_kwarg_list,
        chunksize=1,
        max_workers=max_workers,
        total=len(batch_seq),
    )

    # Stitch the results back together
    out_names = []

    for output_type in output_types:

        if output_type == "video":
            print("Stitching parallel videos")
            stitched_vid_name = join(vid_dir, vid_name + proc_suffix + vid_ext)
            print(stitched_vid_name)
            with VideoWriter(stitched_vid_name) as stitched_vid:
                for i, vid_name in enumerate(
                    [tup[writers_vid_idx] for tup in parallel_writer_fullfiles]
                ):
                    print(f"Stitching video {i}")
                    with VideoReader(vid_name) as cheffed_vid:
                        for frame in cheffed_vid:
                            stitched_vid.append(frame)

            # Check n_expected_frames matches. If not, something is wrong
            print(stitched_vid_name)
            n_stitched_frames = count_frames(stitched_vid_name)
            if not (n_expected_frames == n_stitched_frames):
                raise RuntimeError(
                    f"Frame number in processed video ({n_stitched_frames}) "
                    + f"doesnt match expected ({n_expected_frames}). Something went wrong!"
                )

            out_names.append(stitched_vid_name)

        elif output_type == "arrays":
            print("Stitching arrays")
            stitched_npz_name = join(vid_dir, vid_name + proc_suffix + ".npz")

            # Dictionary to hold stitched np arrays
            results = {}
            counters = {}
            cheffed_npzs = [
                np.load(arr_name)
                for arr_name in [
                    tup[writers_npz_idx] for tup in parallel_writer_fullfiles
                ]
            ]
            for iNpz, npz in enumerate(cheffed_npzs):

                # each npz contains some set of output var, each of which is a key in the npz.
                # For each output var, append it to its stitched array.
                for k in npz.keys():
                    if k not in results:
                        if type(npz[k]) is not np.ndarray:
                            raise ValueError(
                                f"Expected output {k} from function to have type np.ndarray but was {type(npz[k])} instead!"
                            )
                        results[k] = np.zeros((n_expected_frames, *npz[k].shape[1:]))
                        counters[k] = 0
                    len_ = len(npz[k])
                    results[k][(counters[k]) : (counters[k] + len_), ...] = npz[k]
                    counters[k] += len_
            np.savez(stitched_npz_name, **results)

            # Close the npz files
            for npz in cheffed_npzs:
                npz.close()

            out_names.append(stitched_npz_name)

    # Remove the tmp data
    if remove_chunks:
        for file in listdir(tmp_dir):
            remove(join(tmp_dir, file))

    return out_names



from os.path import dirname, isfile, join, realpath
from time import sleep

import numpy as np
from videochef.util import count_frames
from videochef.viz import peri_event_vid

TEST_DATA_DIR = join(dirname(realpath(__file__)), "../test_data")


def test_peri_event_vid(tmp_path):

    # Set up
    path = str(TEST_DATA_DIR)
    assert isfile(join(path, "labeled_frames.avi"))
    test_movie = join(path, "labeled_frames.avi")

    # Prep the peri-event fr lists
    peri_evt_frames_list = [np.arange(i, i + 10) for i in np.arange(0, 300, 30)]
    # out_vid = join(tmp_path, "labeled_frames_PERIEVT.mp4")
    out_vid = join(TEST_DATA_DIR, "labeled_frames_PERIEVT.mp4")

    # Make the peri-event vid
    peri_event_vid(
        test_movie,
        out_vid,
        peri_evt_frames_list,
        event_frame_num_in_vid=5,
        event_end_frame_num_in_vid=7,
        out_fps=10,
        overwrite=True,
    )

    # Not sure why there's a race condition here, but there is
    sleep(2)
    nframes = count_frames(out_vid)
    assert nframes == 10


# if __name__ == '__main__':
#     test_movie = join('../test_data', 'labeled_frames.avi')
#     peri_evt_frames_list = [np.arange(i,i+10) for i in np.arange(0,300,30)]
#     out_vid = './tmp_perievt/labeled_frames_PERIEVT.avi'
#     peri_event_vid(
#         test_movie,
#         out_vid,
#         peri_evt_frames_list,
#         out_fps=10,
#         overwrite=True,
#     )
#     print('done')



from os.path import dirname, isfile, join, realpath

import numpy as np
from videochef.chef import video_chef
from videochef.io import VideoReader, VideoWriter
from videochef.util import (
    count_frames,
    dummy_process,
    dummy_process_arrays,
    dummy_process_vid_and_arrays,
)

TEST_DATA_DIR = join(dirname(realpath(__file__)), "../test_data")
# FIXTURE_DIR = join(dirname(realpath(__file__)), '../test_data')


# TODO: make test for global kwargs + frame by frame kwargs


def test_compare_serial_and_cheffed_labeled_avi(tmp_path):

    # Set up
    path = str(TEST_DATA_DIR)
    assert isfile(join(path, "labeled_frames.avi"))
    test_movie = join(path, "labeled_frames.avi")

    # First, process it serially
    print("Processing serially...")
    serial_vid_name = join(tmp_path, "proc.avi")
    with VideoReader(test_movie) as raw_vid, VideoWriter(serial_vid_name) as serial_vid:
        for frame in raw_vid:
            out = dummy_process(frame)  # returns a tuple of (frame,)
            serial_vid.append(out[0])

    # Then in parallel
    stitched_vid_name = video_chef(
        dummy_process,
        test_movie,
        tmp_dir=tmp_path,
        remove_chunks=False,
        output_types=["video"],
        overwrite_tmp=False,
    )  # must pass overwrite_tmp=false here, or it will overwrite the serial version!
    print(stitched_vid_name)
    stitched_vid_name = stitched_vid_name[
        0
    ]  # video_chef returns a list of output paths, but we only have one output

    # Assert equal num frames
    assert count_frames(serial_vid_name) == count_frames(stitched_vid_name)

    # Check all frames are equal
    equal_frames = np.zeros(count_frames(serial_vid_name))
    non_equal_counter = 0
    with VideoReader(serial_vid_name) as serial_vid, VideoReader(
        stitched_vid_name
    ) as stitched_vid:
        for iFrame, (serial_frame, stitched_frame) in enumerate(
            zip(serial_vid, stitched_vid)
        ):
            if not np.all(serial_frame == stitched_frame):
                print(f"Frame {iFrame} not equal in serial and stitched videos")
                non_equal_counter += 1
            else:
                equal_frames[iFrame] = 1
    assert non_equal_counter == 0


def test_array_func(tmp_path):

    # Set up
    path = str(TEST_DATA_DIR)
    assert isfile(join(path, "labeled_frames.avi"))
    test_movie = join(path, "labeled_frames.avi")

    # Do the processing with a func that returns a dict of values
    # will return: {'avg': avg, 'min': _min, 'max': _max, 'four_middle_px': four_middle_px}
    stitched_npz_name = video_chef(
        dummy_process_arrays,
        test_movie,
        tmp_dir=tmp_path,
        output_types=["arrays"],
        remove_chunks=False,
    )

    stitched_npz_name = stitched_npz_name[
        0
    ]  # video_chef returns a list of names, but we only have one here
    npz = np.load(stitched_npz_name)
    nframes = count_frames(test_movie)
    assert nframes == len(npz["avg"])
    assert npz["four_middle_px"].shape == (nframes, 2, 2)


def test_vid_and_array(tmp_path):

    # Set up
    path = str(TEST_DATA_DIR)
    assert isfile(join(path, "labeled_frames.avi"))
    test_movie = join(path, "labeled_frames.avi")

    # Do the processing with a func that returns a dict of values
    # will return: {'avg': avg, 'min': _min, 'max': _max, 'four_middle_px': four_middle_px}
    stitched_names = video_chef(
        dummy_process_vid_and_arrays,
        test_movie,
        tmp_dir=tmp_path,
        output_types=["video", "arrays"],
        remove_chunks=False,
        overwrite_tmp=False,
    )

    stitched_vid_name, stitched_npz_name = tuple(
        stitched_names
    )  # video_chef returns a list of names

    npz = np.load(stitched_npz_name)
    nframes = count_frames(test_movie)
    assert nframes == len(npz["avg"])
    assert npz["four_middle_px"].shape == (nframes, 2, 2)

    assert count_frames(test_movie) == count_frames(stitched_vid_name)



import datetime
import os
import pdb
import subprocess
import warnings

import av
import numpy as np


class VideoWriter:
    """A simple, ffmpeg-based video writer. Will try to infer grayscale vs. color (use RGB!), and act accordingly.
    Inputs should be (nframes) x (w) x (h) x (RGB).

    Low-level ffmpeg options that you might want to use are:

        preset {str} -- ffmpeg preset to use. (default: {'veryfast'})
        stderrfile {str} -- path to send ffmpeg stderr to, if any. Can't capture stderr programatically due to how we write the frames one at a time.
        crf {int} -- quality factor. Normal default is 23, here we boost it a bit (lower is better), because science. (default: {10})
        ffmpeg_loglevel {str} -- loglevel for ffmpeg. Useful if you get weird bugs and want to debug.

    VideoWriter will try to detect B/W vs color, and set the pixel format and codec accordingly.
    """

    def __init__(self, file_name, verbose=False, **write_frames_options):
        self.pipe = None
        self.file_name = file_name
        self.write_frame_options = write_frames_options
        self.verbose = verbose

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.pipe is not None:
            self.pipe.stdin.close()
            exitcode = (
                self.pipe.wait()
            )  # wait for ffmpeg to finish so the video is ready to go on the next line of code; otherwise run into weird conditions where video isn't saved to disk yet
            if exitcode != 0:
                print("ffmpeg exited with error code: ", exitcode)

    def append(self, frames):
        """Write frames to the file.
        Will convert frames into an array (nframes) x (w) x (h) x color, by inference.
        Frames should have color as final dimension, if at all.

        Arguments:
            frames {[type]} -- [description]
        """

        # Infer whether or not the frames have color
        # Will fail if the video is 3 px tall...seems unlikely.
        is_color = (len(frames.shape) >= 3) and (frames.shape[-1] == 3)

        # Re-shape the array accordingly
        if is_color and (len(frames.shape) == 3):
            # If color but only one frame, add a singleton dim in front
            frames = frames[None]
        elif is_color and (len(frames.shape) == 4):
            pass
        elif len(frames.shape) == 2:
            # If not color and only one frame, add a singleton dimension in front
            frames = frames[None]

        # Set default options if video is in color or gray.
        if is_color:
            pixel_format = self.write_frame_options.pop(
                "pixel_format", "yuv420p"
            )  # output pixel format
            codec = self.write_frame_options.pop("codec", "libx264")
        else:
            pixel_format = self.write_frame_options.pop("pixel_format", "gray8")
            codec = self.write_frame_options.pop("codec", "ffv1")

        if "ffmpeg_loglevel" not in self.write_frame_options:
            if self.verbose:
                ffmpeg_loglevel = "info"
            else:
                ffmpeg_loglevel = "error"

        self.pipe = write_frames(
            self.file_name,
            frames,
            pipe=self.pipe,
            pixel_format=pixel_format,
            codec=codec,
            ffmpeg_loglevel=ffmpeg_loglevel,
            **self.write_frame_options,
        )


class VideoReader:
    def __init__(self, file_name, frame_ixs=None, reporter_val=None, mp4_to_gray=False):
        """
        A simple video reader that uses PyAV to read in frames from a video file.

        Parameters
        ----------
        file_name : str
            Full path to the video file.

        frame_ixs : list or array, optional
            List of frame indices to read in. If None, will read in all frames. (default: None)

        reporter_val : int, optional
            If not None, will print out the frame number being read in. Used for debugging. (default: None)

        mp4_to_gray : bool, optional
            If True, will convert mp4s to grayscale. (default: False)

        """

        self.file_name = file_name
        self.file_ext = os.path.splitext(self.file_name)[1]
        self.reporter_val = reporter_val
        self.mp4_to_gray = mp4_to_gray
        self.current_frame_ix = -1

        if frame_ixs is not None:
            if isinstance(frame_ixs, int):
                frame_ixs = [frame_ixs]
            if isinstance(frame_ixs, list):
                frame_ixs = np.array(frame_ixs)
            if frame_ixs.ndim > 1:
                warnings.warn("frame_ixs should be 1D, not 2D. Flattening it.")
                frame_ixs = frame_ixs.ravel()
            self.frame_ixs = np.sort(frame_ixs)
            self.final_frame_ix = int(np.max(frame_ixs))
        else:
            self.frame_ixs = None
            self.final_frame_ix = None

        if self.reporter_val is not None:
            self.reporter_val = int(reporter_val)

        # TODO: calculate threshold above which to use fast_seek
        self.fask_seek_threshold = 1000  # nframes

    def __enter__(self):
        self.reader = av.open(self.file_name, "r")
        self.reader.streams.video[0].thread_type = "AUTO"
        self.codec = self.reader.streams.video[0].name
        self.pix_fmt = self.reader.streams.video[0].format.name
        self.rate = self.reader.streams.video[0].average_rate
        self.time_base = self.reader.streams.video[0].time_base
        self.start_time = self.reader.streams.video[0].start_time

        # Check frame_ixs is within an ok range
        if self.frame_ixs is not None:
            assert self.frame_ixs.max() < self.reader.streams.video[0].frames
        else:
            # Make frame_ixs if it's not already made
            self.frame_ixs = np.arange(self.reader.streams.video[0].frames)

        # Return a generator object that will decode the frames into np arrays
        return self.frame_gen(self.reader.decode(video=0))

    def _plain_frame_gen(self):
        for packet in self.reader.demux(
            self.reader.streams.video[0]
        ):  # TODO: check if packet has time attribute? If so, can use packet time instead of frame time here so we dont hvae to decode and it'll speed it up?
            for frame in packet.decode():
                yield frame

    def _precise_seek(self, frame_num_to_seek):
        """See https://github.com/PyAV-Org/PyAV/blob/b65f5a9f93144d2eadbb3e460bb45d869f5ce2fe/scratchpad/second_seek_example.py"""
        target_sec = frame_num_to_seek * 1 / self.rate
        target_pts = int(target_sec / self.time_base) + self.start_time
        self.reader.seek(target_pts, stream=self.reader.streams.video[0])
        for frame in self._plain_frame_gen():
            # frame_num = np.round(frame.time * self.rate, 0)  # this seems more reliable
            # print(frame_num, frame_num_to_seek)
            # if frame_num >= frame_num_to_seek:
            if frame.pts >= target_pts:
                # return frame
                break  # instead of returning the frame, we'll just break and let the frame_gen yield it

    def frame_gen(self, reader):
        """Yield frames from the video file.
        Will fast-seek to the next frame if the difference between the current frame and the next frame is greater than the fask_seek_threshold.

        Parameters
        ----------
        reader : av container
            PyAV container object that contains the video file.
        """
        for frame_ix in self.frame_ixs:

            # Fask-seek to the next frame if needed
            if (frame_ix - self.current_frame_ix) > self.fask_seek_threshold:
                self._precise_seek(
                    frame_ix - 1
                )  # seek to the frame before the one we want, so that the next call to next(reader) will give us the frame we want

            # Else just read frames normally until we get to the next desired frame
            elif (frame_ix - self.current_frame_ix) > 0:
                for _ in range(frame_ix - self.current_frame_ix - 1):
                    next(reader)

            # Yield the frame
            self.current_frame_ix = frame_ix
            yield self.convert_frame_to_np(next(reader))

            if self.reporter_val is not None:
                print(f"read frame {frame_ix} from reader {self.reporter_val}")

    def convert_frame_to_np(self, frame):
        if self.file_ext == ".avi" and self.codec == "ffv1" and "gray" in self.pix_fmt:
            return frame.to_ndarray()
        elif self.file_ext == ".avi" and self.codec == "h264" and "yuv" in self.pix_fmt:
            return np.array(frame.to_image())
        elif (
            (
                self.file_ext == ".mp4"
                and self.codec == "h264"
                and (
                    self.pix_fmt == "yuvj420p"
                    or self.pix_fmt == "yuv420p"
                    or self.pix_fmt == "yuv444p"
                )
            )
            or (
                self.file_ext == ".mp4"
                and (self.codec == "h265" or self.codec == "hevc")
                and (self.pix_fmt == "yuvj420p" or self.pix_fmt == "yuv420p")
            )
            or (
                self.file_ext == ".mp4"
                and self.codec == "mpeg4"
                and self.pix_fmt == "yuv420p"
            )
        ):
            if self.mp4_to_gray:
                return np.array(frame.to_image())[:, :, 0]
            else:
                return np.array(frame.to_image())
        else:
            raise NotImplementedError(
                f"Video format {self.file_ext} with codec {self.codec} and px fmt {self.pix_fmt} is not supported yet."
            )

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.reader.close()


def write_frames(
    filename,
    frames,
    threads=6,
    fps=30,
    crf=10,
    pixel_format="gray8",
    codec="ffv1",
    pipe=None,
    slices=24,
    slicecrc=1,
    preset="veryfast",
    stderrfile=None,
    ffmpeg_loglevel="info",
):
    """Use ffmpeg to write frames into a movie

    Arguments:
        filename {str} -- full filename of video to write
        frames {iterable of arrays} -- the frames to write. Should be numpy arrays (probably uint8's)

    Keyword Arguments:
        threads {int} -- n threads for ffmpeg (default: {6})
        fps {int} -- desired output fps (default: {30})
        crf {int} -- quality factor. Normal default is 23, here we boost it a bit (lower is better), because science. (default: {10})
        pixel_format {str} -- output pixel format. For quicktime playable movies, use "yuv420p" and pass 3D frames (). (default: {'gray8'})
        codec {str} -- video codec to use. For color, prefer "libx264" (default: {'ffv1'})
        pipe {[type]} -- internal use (default: {None})
        slices {int} -- number of ffmpeg slices (default: {24})
        slicecrc {int} -- ? (default: {1})
        preset {str} -- ffmpeg preset to use. (default: {'veryfast'})
        stderrfile {str} -- path to send ffmpeg stderr to, if any. Can't capture stderr programatically due to how we write the frames one at a time.
        ffmpeg_loglevel {str} -- loglevel for ffmpeg. Info is their default.

    Returns:
        [type] -- [description]
    """

    frame_size = "{0:d}x{1:d}".format(frames.shape[2], frames.shape[1])

    if pixel_format == "gray8" or "rgb" in pixel_format:
        command = [
            "ffmpeg",
            "-y",
            "-loglevel",
            ffmpeg_loglevel,
            "-framerate",
            str(fps),
            "-f",
            "rawvideo",
            "-s",
            frame_size,
            "-pix_fmt",
            pixel_format,
            "-i",
            "-",
            "-an",
            "-crf",
            str(crf),
            "-vcodec",
            codec,
            "-preset",
            preset,
            "-threads",
            str(threads),
            "-slices",
            str(slices),
            "-slicecrc",
            str(slicecrc),
            "-r",
            str(fps),
            filename,
        ]
    elif "yuv" in pixel_format:
        # assume user is passing rgb (unlikely to be passing yuv formatted arrays)
        command = [
            "ffmpeg",
            "-y",
            "-loglevel",
            "info",
            "-framerate",
            str(fps),
            "-f",
            "rawvideo",
            "-s",
            frame_size,
            "-pix_fmt",
            "rgb24",  # user will pass input as rgb...
            "-i",
            "-",
            "-an",
            "-crf",
            str(crf),
            "-c:v",
            "libx264",
            "-pix_fmt",
            pixel_format,  # ...but output will be converted to yuv
            "-preset",
            preset,
            "-threads",
            str(threads),
            "-slices",
            str(slices),
            "-slicecrc",
            str(slicecrc),
            "-r",
            str(fps),
            filename,
        ]

    # Prep error logging
    if stderrfile is not None:
        errorfile = open(stderrfile, "a")
    else:
        errorfile = None

    # Write the frames
    if not pipe:
        pipe = subprocess.Popen(command, stdin=subprocess.PIPE, stderr=errorfile)
    dtype = np.uint16 if pixel_format.startswith("gray16") else np.uint8
    for i in range(frames.shape[0]):
        pipe.stdin.write(frames[i, ...].astype(dtype).tobytes())

    # Close file, if open
    if errorfile is not None:
        errorfile.close()

    return pipe


def read_frames(
    filename,
    frames,
    threads=6,
    fps=30,
    frames_is_timestamp=False,
    pixel_format="gray16",
    frame_size=(640, 576),
    slices=24,
    slicecrc=1,
    get_cmd=False,
):
    """
    Reads in frames from the .mp4/.avi file using a pipe from ffmpeg.
    Args:
        filename (str): filename to get frames from
        frames (list or 1d numpy array): list of frames to grab
        threads (int): number of threads to use for decode
        fps (int): frame rate of camera in Hz
        pixel_format (str): ffmpeg pixel format of data
        frame_size (str): wxh frame size in pixels
        slices (int): number of slices to use for decode
        slicecrc (int): check integrity of slices
    Returns:
        3d numpy array:  frames x h x w
    """

    if frames_is_timestamp:
        start_time = str(datetime.timedelta(seconds=frames[0]))
    else:
        start_time = str(datetime.timedelta(seconds=frames[0] / fps))

    command = [
        "ffmpeg",
        "-loglevel",
        "info",
        "-vsync",
        "0",
        "-ss",
        start_time,
        "-i",
        filename,
        "-vframes",
        str(len(frames)),
        "-f",
        "image2pipe",
        "-s",
        "{:d}x{:d}".format(frame_size[0], frame_size[1]),
        "-pix_fmt",
        pixel_format,
        "-threads",
        str(threads),
        "-slices",
        str(slices),
        "-slicecrc",
        str(slicecrc),
        "-vcodec",
        "rawvideo",
        "-",
    ]

    if get_cmd:
        return command

    pipe = subprocess.Popen(command, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = pipe.communicate()
    if err:
        print("error", err)
        return None

    dtype = "uint16" if "16" in pixel_format else "uint8"
    video = np.frombuffer(out, dtype=dtype).reshape(
        (len(frames), frame_size[1], frame_size[0])
    )
    return video



from contextlib import ExitStack
from os.path import exists

import cv2
import numpy as np
from tqdm.auto import tqdm

from videochef.io import VideoReader, VideoWriter


def add_titles(
    panels, titles, text_origin=(5, 18), text_size=0.4, text_color=(255, 255, 255)
):
    """Helper function to add titles to a set of video panels

    Arguments:
        panels {list of np.array} -- a list of 2d arrays
        titles {list of str} -- a list of the titles for each panel

    Keyword Arguments:
        text_origin {tuple} -- where to place the text on the panel (default: {(5,18)})
        text_size {float} -- cv2 text size (default: {0.4})
        text_color {tuple} -- cv2 text color (default: {(255,255,255)})

    Returns:
        panels -- panels with titles appended
    """

    return [
        cv2.putText(
            panel.copy(),
            title,
            text_origin,
            cv2.FONT_HERSHEY_SIMPLEX,
            text_size,
            text_color,
            1,
            cv2.LINE_AA,
        )
        for panel, title in zip(panels, titles)
    ]


def peri_event_vid(
    in_vid_name,
    out_vid_name,
    peri_evt_frames_list,
    event_frame_num_in_vid=None,
    event_end_frame_num_in_vid=None,
    traces=None,
    trace_to_vid_fs_ratio=None,
    out_fps=12,
    overwrite=False,
    verbose=False,
    **write_frame_kwargs,
):
    """Create a tiled gallery of peri-event videos

    Arguments:
        in_vid_name {path} -- video to read
        out_vid_name {path} -- video to write
        peri_evt_frames_list {list of iterables} -- peri-event frame lists for each event
        event_frame_num_in_vid {int} -- if passed, mark before/after each event onset with a hollow/filled circle. Otherwise not shown.
        event_end_frame_num_in_vid {int} -- if passed, change the circle back to hollow at this frame (after the event is over).

    Keyword Arguments:
        traces {array} -- TODO: allow user to display sync'd traces (default: {None})
        trace_to_vid_fs_ratio {[type]} -- TODO: ^^ (default: {None})
        out_fps {int} -- desired fps of written video. Often nice to slow it down. (default: {12})
        overwrite {bool} -- if True, overwrites any existing video at out_vid_name (default: {False})

    Low-level ffmpeg options that can also be passed as kwargs are:
        preset {str} -- ffmpeg preset to use. (default: {'veryfast'})
        stderrfile {str} -- path to send ffmpeg stderr to, if any. Can't capture stderr programatically due to how we write the frames one at a time.
        crf {int} -- quality factor. Normal default is 23, here we boost it a bit (lower is better), because science. (default: {10})
        ffmpeg_loglevel {str} -- loglevel for ffmpeg. Useful if you get weird bugs and want to debug.

    Returns: n/a

    Dummy example:
        # makes peri-event vids from 0.5 sec before to 1.5 sec after the event's frame
        movie_path = '/path/to/your/movie.mp4'  # at least ~ 1100 frames long for this example
        fps = 30
        overwrite = False
        window = (-0.5, 1.5)
        event_idx_in_vid = np.arange(0,1000,100)
        peristim_frames_list = [np.arange(idx+window[0]*vid_fps, idx+window[1]*vid_fps) for idx in event_idx_in_vid]
        out_vid_name = movie_path.replace('movie', 'perievent_movie')
        try:
            videochef.viz.peri_event_vid(
                movie_path,
                out_vid_name,
                peri_evt_frames_list=peristim_frames_list,
                event_frame_num_in_vid=(0 - window[0])*vid_fps,
                out_fps=fps/5,  # slow down the output
                overwrite=overwrite,
            )
        except BrokenPipeError:
            warnings.warn('Encountered broken pipe, skipping...')

    """

    if traces is not None:
        raise NotImplementedError()

    # Plan layout to be as close to square as possible
    num_stims = len(peri_evt_frames_list)
    nrows = np.ceil(np.sqrt(num_stims)).astype("int")

    # Skip already existing vids
    if exists(out_vid_name) and not overwrite:
        return None

    # libx264 is a better verison of h264.
    # yuv420 is open-able by quicktime and most video players.
    # NB, user should still pass in RGB frames -- it will be converted by ffmpeg into yuv (see videochef.io.write_frames()).
    waiting_writer = VideoWriter(
        out_vid_name,
        fps=out_fps,
        codec="libx264",
        pixel_format="yuv420p",
        verbose=verbose,
        **write_frame_kwargs,
    )
    with ExitStack() as stack:

        # Get video readers (and they seek to exactly the right place on this line!)
        if verbose:
            print("Getting readers...")
        peristim_readers = [
            stack.enter_context(
                VideoReader(
                    in_vid_name,
                    np.array(frames),
                )
            )
            for frames in peri_evt_frames_list
        ]

        if verbose:
            print("getting writer")
        writer = stack.enter_context(waiting_writer)

        # Iterate through all readers in parallel ((frame0 stim0, frame0 stim1,...), (frame1 stim0, frame1 stim1,...), ...)
        # which will get condensed across stims into bigframe 0, bigframe 1, bigframe 2,...
        # ix counts aligned frame num (ie starts at 0)
        for ix, frames in tqdm(
            enumerate(zip(*peristim_readers)), total=len(peri_evt_frames_list[0])
        ):
            all_rows = []

            # Add each mini-frame to the big frame
            panels, titles = [], []  # empty lists for frame

            ii = 0  # counter for layout
            for iReader, frame in enumerate(frames):

                # Convert frame to RGB if it isn't already
                if frame.ndim == 2:
                    frame = np.stack(3 * (frame,), axis=-1)
                elif frame.ndim != 3:
                    raise ValueError(
                        "Expected 3D frames but got frames of shape {frame.shape}"
                    )

                # Append a mark for whether event fr has passed or not
                if event_frame_num_in_vid is not None:
                    fill = 1
                    if (
                        (ix >= event_frame_num_in_vid) 
                        and (
                            (event_end_frame_num_in_vid is None)
                            or (ix < event_end_frame_num_in_vid)
                        )
                    ):
                        fill = -1  # for some reason, -1 fills it

                    frame = cv2.circle(frame, (18, 96), 15, (0, 255, 0), fill)

                # Mark traces if requested
                # TODO: have to mark real-time on the trace by tracking the right xval. Boundary conditions make it hard. pure midpoint only works when neither boundary is limiting.
                #                 trace_start_idx = int(max(0, (ix*trace_to_vid_fs_ratio - 200)))
                #                 trace_end_idx = int(min(traces.shape[1], (ix*trace_to_vid_fs_ratio + 200)))
                #                 slice_to_show = slice(trace_start_idx, trace_end_idx, 2)
                #                 if traces is not None:
                #                     vals = traces[iReader, slice_to_show]
                #                     vals_on_frame_x = np.arange(len(vals))
                #                     vals_on_frame_y = 150 + -1*vals*20
                #                     pts = np.column_stack([vals_on_frame_x, vals_on_frame_y]).round().astype(np.int32)
                #                     frame = cv2.polylines(frame, [pts], False, (0,255,0))
                #                     midpoint_idx = len(pts)//2
                #                     frame = cv2.circle(frame, pts[midpoint_idx], 3, (255,0,0), fill)

                # Add to frame
                panels.append(frame.astype("uint8"))
                titles.append(f"Instance: {iReader}; fr: {ix}")

                # If at end of row, or end of rectangle, fill out big frame
                if ii == (nrows - 1):
                    panels = add_titles(
                        panels, titles, text_size=1, text_origin=(18, 32)
                    )
                    #                     panels = add_titles(panels, odor_infos, text_size=1, text_origin=(18,64))
                    all_rows.append(panels)
                    ii = 0
                    panels, titles = [], []
                elif iReader == (len(frames) - 1):
                    panels = add_titles(
                        panels, titles, text_size=1, text_origin=(18, 32)
                    )
                    all_rows.append(panels)
                    ii = 0
                else:
                    ii += 1

            # Stack the individual panels into rows
            stacked_rows = [np.hstack(panels) for panels in all_rows]

            # Deal with any non-full rows
            max_row = np.argmax([row.size for row in stacked_rows])
            for iRow in range(len(stacked_rows)):
                diff = stacked_rows[max_row].shape[1] - stacked_rows[iRow].shape[1]
                if diff != 0:
                    stacked_rows[iRow] = np.pad(
                        stacked_rows[iRow],
                        ((0, 0), (0, diff), (0, 0)),
                        constant_values=0,
                    )

            # Stack the rows (now all identical shapes)
            big_frame = np.vstack(stacked_rows).astype("uint8")

            # Height and width must be divisible by two, at least for libx264
            for dim in range(big_frame.ndim):
                big_shape = big_frame.shape
                if (dim == 0 or dim == 1) and big_shape[dim] % 2 == 1:
                    zeros_shape = list(big_shape)
                    zeros_shape[dim] = 1
                    big_frame = np.concatenate(
                        [big_frame, np.zeros(zeros_shape)], axis=dim
                    )

            if verbose:
                print(f"Writing frame {ix}...")
                print(f"Frame shape: {big_frame.shape}")
            writer.append(big_frame)

    return



from os.path import dirname, isfile, join, realpath

import matplotlib.pyplot as plt
import numpy as np
import videochef as vc
from videochef.io import VideoReader

TEST_DATA_DIR = join(dirname(realpath(__file__)), "../test_data")


def test_make_batch_sequence():
    nfr = 100
    chunk_size = 10
    overlap = 0
    offset = 0

    seq = vc.util.make_batch_sequence(nfr, chunk_size, overlap, offset=offset)
    assert seq[0] == range(0, 10)

    seq = vc.util.make_batch_sequence(nfr, chunk_size, overlap, offset=5)
    assert seq[0] == range(5, 15)
    assert seq[-1] == range(95, 100)

    seq = vc.util.make_batch_sequence(nfr, chunk_size, 5, offset=5)
    assert seq[0] == range(5, 15)
    assert seq[1] == range(10, 20)
    assert seq[-2] == range(85, 95)
    assert seq[-1] == range(90, 100)


def test_count_frames_avi():

    # Set up
    path = str(TEST_DATA_DIR)
    assert isfile(join(path, "labeled_frames.avi"))
    test_movie = join(path, "labeled_frames.avi")

    nfr = vc.util.count_frames(test_movie)
    assert nfr == 7148


def test_count_frames_mp4():

    # Set up
    path = str(TEST_DATA_DIR)
    assert isfile(join(path, "labeled_frames_mark_2221.mp4"))
    test_movie = join(path, "labeled_frames_mark_2221.mp4")

    nfr = vc.util.count_frames(test_movie)
    assert nfr == 10000


def test_precise_seek_avi():

    # Set up
    path = str(TEST_DATA_DIR)
    assert isfile(join(path, "labeled_frames.avi"))
    test_movie = join(path, "labeled_frames.avi")

    frame_to_check = 300
    with VideoReader(test_movie, np.array([frame_to_check])) as vid:
        for frame in vid:
            assert np.sum(frame[0, :]) == frame_to_check


def test_precise_seek_mp4_h264():

    # Set up
    path = str(TEST_DATA_DIR)
    assert isfile(join(path, "labeled_frames_mark_2221.mp4"))
    test_movie = join(path, "labeled_frames_mark_2221.mp4")

    frame_to_check = 2221
    with VideoReader(
        test_movie, np.array([frame_to_check]), mp4_to_gray=True, reporter_val=1
    ) as vid:
        for i, frame in enumerate(vid):
            # plt.imshow(frame)
            # plt.show()
            marker_pix_val = 255
            buffer_val = 15  # h264 compression means not all px at exactly 255, give it some buffer space.
            n_marked_rows = 10
            n_cols = 400
            assert np.sum(frame[0:10, :]) > (
                (marker_pix_val - buffer_val) * n_marked_rows * n_cols
            )


def test_precise_seek_mp4_h265():

    # Set up
    path = str(TEST_DATA_DIR)
    test_movie = join(path, "labeled_frames_mark_2221_H265.mp4")
    assert isfile(test_movie)

    frame_to_check = 2221
    with VideoReader(test_movie, np.array([frame_to_check]), mp4_to_gray=True) as vid:
        for frame in vid:
            marker_pix_val = 255
            buffer_val = 15  # h264 compression means not all px at exactly 255, give it some buffer space.
            n_marked_rows = 10
            n_cols = 400
            assert np.sum(frame[0:10, :]) > (
                (marker_pix_val - buffer_val) * n_marked_rows * n_cols
            )


def test_precise_seek_mp4_h264_multiple_frames():

    # Set up
    path = str(TEST_DATA_DIR)
    test_movie = join(path, "labeled_frames_mark_2221.mp4")
    assert isfile(test_movie)

    # frames_to_check = [0, 10, 100, 110, 120, 2000, 8000, 8100, 9200]
    # frames_to_check = [0, 500, 1000, 1500, 3500]
    frames_to_check = [0, 1000, 3000, 3100, 3101, 6000]
    # frames_to_check = [0, 1001, 3000, 5000, 7000, 9000, 9999]
    frames = []
    with VideoReader(test_movie, np.array([frames_to_check]), mp4_to_gray=True) as vid:
        for i, frame in enumerate(vid):
            frames.append(frame)

    plt.figure()
    for i, frame in enumerate(frames):
        plt.subplot(3, 3, i + 1)
        plt.imshow(frame)
    plt.show()
    assert len(frames) == len(frames_to_check)



import av
import numpy as np


def make_batch_sequence(nframes, chunk_size, overlap, offset=0, step=1):
    """
    Generates batches used to chunk videos prior to extraction.
    Parameters
    ----------
    nframes (int): total number of frames
    chunk_size (int): desired chunk size
    overlap (int): number of overlapping frames
    offset (int): frame offset
    Returns
    -------
    Returns list of batches
    """

    if overlap >= chunk_size:
        raise ValueError("Overlap must be less than chunk size")

    seq = range(offset, nframes, step)
    out = []
    for i in range(0, len(seq) - overlap, chunk_size - overlap):
        out.append(seq[i : i + chunk_size])
    return out


def unwrap_dictionary(dict_):
    """Take a dictionary of form {'a': vals, 'b': vals, ...} and convert to a list of form [{'a': val1, 'b': val2, ...}, {'a': val2, 'b': val2, ...}, ...]
        Vals must all be same length!

    Arguments:
        dict_ {[type]} -- [description]
    """

    # Transform kwarg dict from {'a': vals, 'b': vals, ...} into [{'a': val1, 'b': val1,...}, {'a': val2, 'b': val2,...}, ...]
    tmp_lists = (
        []
    )  # Intermdiate step: first unpack each value for each dict [[{'a':val1, 'a': val2}], [{'b': val1}, {'b':val2}], ...]
    for key in dict_.keys():
        tmp_lists.append([{key: v} for v in dict_[key]])

    # Then put into desired final form
    final_list = []
    for dicts in zip(*tmp_lists):
        final_list.append({k: v for d in dicts for k, v in d.items()})

    return final_list


def dummy_process(frame):
    # Normalize the frame's pixel vals around half max brightness (just something stupid that isn't instant)
    avg = np.mean(frame)
    std = np.std(frame)
    frame = ((frame - avg) / std + 255 / 2).astype("uint8")
    return (frame,)


def dummy_process_arrays(frame):
    # Get the min, max, and mean of each frame
    # Also get the 4 middle pixels, just to test the ability to store >1-D arrays
    avg = np.mean(frame)
    _min = np.min(frame)
    _max = np.max(frame)
    midx = frame.shape[0] // 2
    midy = frame.shape[1] // 2
    four_middle_px = frame[midx - 1 : midx + 1, midy - 1 : midy + 1]

    # return a dictionary with the results in it
    out = {"avg": avg, "min": _min, "max": _max, "four_middle_px": four_middle_px}

    return (out,)


def dummy_process_vid_and_arrays(frame):
    """Return (frame, out_dict)"""

    # Normalize the frame's pixel vals around half max brightness (just something stupid that isn't instant)
    avg = np.mean(frame)
    std = np.std(frame)
    frame = ((frame - avg) / std + 255 / 2).astype("uint8")

    # Get the min, max, and mean of each frame
    # Also get the 4 middle pixels, just to test the ability to store >1-D arrays
    avg = np.mean(frame)
    _min = np.min(frame)
    _max = np.max(frame)
    midx = frame.shape[0] // 2
    midy = frame.shape[1] // 2
    four_middle_px = frame[midx - 1 : midx + 1, midy - 1 : midy + 1]
    out_dict = {"avg": avg, "min": _min, "max": _max, "four_middle_px": four_middle_px}

    return (frame, out_dict)


def count_frames(file_name):
    with av.open(file_name, "r") as reader:
        return reader.streams.video[0].frames



from glob import glob
from os.path import basename, splitext

from setuptools import find_packages, setup

setup(
    name="videochef",
    version="0.2.1",
    license="MIT",
    description="Whip up some fast, embarrassingly parallel video processing",
    author="Jonah Pearl",
    author_email="jonahpearl@g.harvard.edu",
    url="https://github.com/jonahpearl/videochef",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Image Processing",
    ],
    project_urls={
        "Issue Tracker": "https://github.com/jonahpearl/videochef/issues",
    },
    python_requires=">=3.6",
    install_requires=[
        "numpy",
        "click",
        "av",
        "tqdm",
    ],
)
